"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.applySession = void 0;
const cookie_1 = require("cookie");
const nanoid_1 = require("nanoid");
const memory_1 = __importDefault(require("./store/memory"));
const hash = (sess) => JSON.stringify(sess, (key, val) => (key === 'cookie' ? undefined : val));
const commitHead = (res, name, { cookie, id }, encodeFn) => {
    if (res.headersSent)
        return;
    const cookieArr = [
        (0, cookie_1.serialize)(name, encodeFn ? encodeFn(id) : id, {
            path: cookie.path,
            httpOnly: cookie.httpOnly,
            expires: cookie.expires,
            domain: cookie.domain,
            sameSite: cookie.sameSite,
            secure: cookie.secure,
        }),
    ];
    const prevCookies = res.getHeader('set-cookie');
    if (prevCookies) {
        if (Array.isArray(prevCookies))
            cookieArr.push(...prevCookies);
        else
            cookieArr.push(prevCookies);
    }
    res.setHeader('set-cookie', cookieArr);
};
const storeFn = {
    destroy(s, sid) {
        return new Promise((resolve, reject) => {
            const result = s.destroy(sid, (err) => (err ? reject(err) : resolve()));
            if (result && typeof result.then === 'function')
                result.then(resolve, reject);
        });
    },
    get(s, sid) {
        return new Promise((resolve, reject) => {
            const result = s.get(sid, (err, val) => 
            // @ts-ignore: Compat diff
            err ? reject(err) : resolve(val));
            if (result && typeof result.then === 'function')
                result.then(resolve, reject);
        });
    },
    set(s, sid, sess) {
        return new Promise((resolve, reject) => {
            // @ts-ignore: Compat diff
            const result = s.set(sid, sess, (err) => (err ? reject(err) : resolve()));
            if (result && typeof result.then === 'function')
                result.then(resolve, reject);
        });
    },
    touch(s, sid, sess) {
        return new Promise((resolve, reject) => {
            const done = (err) => (err ? reject(err) : resolve());
            // @ts-ignore: Compat diff
            const result = s.touch(sid, sess, done);
            if (result && typeof result.then === 'function')
                result.then(resolve, reject);
        });
    },
};
let memoryStore;
async function applySession(req, res, options = {}) {
    if (req.session)
        return;
    /**
     * Options init
     */
    const name = options.name || 'sid';
    const store = options.store || (memoryStore = memoryStore || new memory_1.default());
    const genId = options.genid || nanoid_1.nanoid;
    const encode = options.encode;
    const decode = options.decode;
    let touchAfter = -1;
    // compat: if rolling is `true`, user might have wanted to touch every time
    // thus defaulting options.touchAfter to 0 instead of -1
    if (options.rolling && !('touchAfter' in options)) {
        console.warn('The use of options.rolling is deprecated. Setting this to `true` without options.touchAfter causes options.touchAfter to be defaulted to `0` (always)');
        touchAfter = 0;
    }
    if (typeof options.touchAfter === 'number') {
        touchAfter = options.touchAfter;
    }
    const autoCommit = typeof options.autoCommit === 'boolean' ? options.autoCommit : true;
    const cookieOpts = (options === null || options === void 0 ? void 0 : options.cookie) || {};
    /**
     * Main
     */
    let isDestroyed = false;
    let isTouched = false;
    const _now = Date.now();
    const resetMaxAge = (session) => {
        isTouched = true;
        session.cookie.expires = new Date(_now + session.cookie.maxAge * 1000);
    };
    let sessId = req.headers && req.headers.cookie ? (0, cookie_1.parse)(req.headers.cookie)[name] : null;
    if (sessId && decode)
        sessId = decode(sessId);
    const loadedSess = sessId ? await storeFn.get(store, sessId) : null;
    let session;
    if (loadedSess) {
        session = loadedSess;
        // Some store return cookie.expires as string, convert it to Date
        if (typeof session.cookie.expires === 'string') {
            session.cookie.expires = new Date(session.cookie.expires);
        }
    }
    else {
        sessId = genId();
        session = {
            cookie: {
                path: cookieOpts.path || '/',
                httpOnly: cookieOpts.httpOnly || true,
                domain: cookieOpts.domain || undefined,
                sameSite: cookieOpts.sameSite,
                secure: cookieOpts.secure || false,
                ...(cookieOpts.maxAge
                    ? {
                        maxAge: cookieOpts.maxAge,
                        expires: new Date(_now + cookieOpts.maxAge * 1000),
                    }
                    : { maxAge: undefined }),
            },
        };
    }
    Object.defineProperties(session, {
        commit: {
            value: async function commit() {
                commitHead(res, name, this, encode);
                await storeFn.set(store, this.id, this);
            },
        },
        destroy: {
            value: async function destroy() {
                isDestroyed = true;
                this.cookie.expires = new Date(1);
                await storeFn.destroy(store, this.id);
                delete req.session;
            },
        },
        isNew: { value: !loadedSess },
        id: { value: sessId },
    });
    // Set to req.session
    req.session = session;
    // Compat with express-session
    req.sessionStore = store;
    // prevSessStr is used to compare the session later
    // for touchability -- that is, we only touch the
    // session if it has changed. This check is used
    // in autoCommit mode only
    let prevHash;
    if (autoCommit) {
        prevHash = hash(session);
    }
    // Extends the expiry of the session
    // if touchAfter is applicable
    if (touchAfter >= 0 && session.cookie.expires) {
        const lastTouchedTime = session.cookie.expires.getTime() - session.cookie.maxAge * 1000;
        if (_now - lastTouchedTime >= touchAfter) {
            resetMaxAge(session);
        }
    }
    // autocommit: We commit the header and save the session automatically
    // by "proxying" res.writeHead and res.end methods. After committing, we
    // call the original res.writeHead and res.end.
    if (autoCommit) {
        const oldWritehead = res.writeHead;
        res.writeHead = function resWriteHeadProxy(...args) {
            if (session.isNew || isTouched || isDestroyed) {
                commitHead(res, name, session, encode);
            }
            return oldWritehead.apply(this, args);
        };
        const oldEnd = res.end;
        res.end = function resEndProxy(...args) {
            const done = () => oldEnd.apply(this, args);
            if (isDestroyed)
                return done();
            if (hash(session) !== prevHash) {
                storeFn.set(store, session.id, session).finally(done);
            }
            else if (isTouched && store.touch) {
                storeFn.touch(store, session.id, session).finally(done);
            }
            else {
                done();
            }
        };
    }
}
exports.applySession = applySession;
